"""Contains the KinClient class to interact with the blockchain"""

from kin_base import Horizon

from .config import ANON_APP_ID, MAX_RECORDS_PER_REQUEST
from . import errors as KinErrors
from .monitors import single_monitor, multi_monitor
from .transactions import SimplifiedTransaction, RawTransaction
from .account import KinAccount
from .blockchain.horizon_models import AccountData
from .blockchain.utils import is_valid_address, is_valid_transaction_hash
from .version import __version__
from .blockchain.environment import Environment

from typing import List, Optional, Union, AsyncGenerator

import logging

logger = logging.getLogger(__name__)


class KinClient:
    """
    The :class:`kin.KinClient` class is the primary interface to the KIN Python SDK based on Kin Blockchain.
    It maintains a connection context with a Horizon node and hides all the specifics of dealing with Kin REST API.
    """

    def __init__(self, environment: Environment):
        """Create a new instance of the KinClient to query the Kin blockchain.

        :param environment: an environment for the client to point to.

        :return: An instance of the KinClient.
        :rtype: KinErrors.KinClient
        """

        self.environment = environment
        self.network = environment.name

        self.horizon = Horizon(environment.horizon_uri)
        logger.info('Kin Client initialized on network {}, horizon endpoint {}'.
                    format(self.network, self.horizon.horizon_uri))

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self) -> None:
        """Close the connection to the horizon server"""
        await self.horizon.close()

    def kin_account(self, seed: str, channel_secret_keys: Optional[List[str]] = None,
                    app_id: Optional[str] = ANON_APP_ID) -> KinAccount:
        """
        Create a new instance of a KinAccount to perform authenticated operations on the blockchain.

        :param seed: The secret seed of the account that will be used
        :param channel_secret_keys: A list of seeds to be used as channels
        :param app_id: the unique id of your app
        :return: An instance of KinAccount
        """

        # Create a new kin account, using self as the KinClient to be used
        return KinAccount(seed, self, channel_secret_keys, app_id)

    async def get_config(self) -> dict:
        """Get system configuration data and online status.

        :return: a dictionary containing the data
        :rtype: dict
        """
        status = {
            'sdk_version': __version__,
            'environment': self.environment.name,
            'horizon': {
                'uri': str(self.horizon.horizon_uri),
                'online': False,
                'error': None,
            },
            'transport': {
                'pool_size': self.horizon._session.connector.limit,
                'num_retries': self.horizon.num_retries,
                'request_timeout': self.horizon._session._timeout.total,
                'backoff_factor': self.horizon.backoff_factor,
            }
        }

        # now check Horizon connection
        try:
            await self.horizon.metrics()
            status['horizon']['online'] = True
        except Exception as e:
            status['horizon']['error'] = repr(e)

        return status

    async def get_minimum_fee(self) -> int:
        """
        Get the current minimum fee acceptable for a tx

        :return: The minimum fee
        """
        params = {'order': 'desc',
                  'limit': 1}
        return (await self.horizon.ledgers(order='desc', limit=1))['_embedded']['records'][0]['base_fee_in_stroops']

    async def get_account_balance(self, address: str) -> float:
        """
        Get the KIN balance of a given account

        :param address: the public address of the account to query
        :return: the balance of the account

        :raises: StellarAddressInvalidError: if the provided address has the wrong format.
        :raises: KinErrors.AccountNotFoundError if the account does not exist.
        """

        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))
        account_data = await self.get_account_data(address)
        for balance in account_data.balances:
            # accounts will always have native asset
            if balance.asset_type == 'native':
                return balance.balance

    async def does_account_exists(self, address: str) -> bool:
        """
        Find out if a given account exists on the blockchain

        :param address: The kin account to query about
        :return: does the account exists on the blockchain

        :raises: KinErrors.StellarAddressInvalidError if the address is not valid.
        """

        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))

        try:
            await self.get_account_balance(address)
            return True
        except KinErrors.AccountNotFoundError:
            return False

    async def get_account_data(self, address: str) -> AccountData:
        """Get account data.

        :param address: the public address of the account to query.

        :return: account data

        :raises: StellarAddressInvalidError: if the provided address has a wrong format.
        :raises: :class:`KinErrors.AccountNotFoundError`: if the account does not exist.
        """
        # TODO: might want to simplify the returning data
        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))

        try:
            acc = await self.horizon.account(address)
            return AccountData(acc, strict=False)
        except Exception as e:
            err = KinErrors.translate_error(e)
            raise KinErrors.AccountNotFoundError(address) if \
                isinstance(err, KinErrors.ResourceNotFoundError) else err

    async def get_transaction_data(self, tx_hash: str, simple: Optional[bool] = True) -> Union[SimplifiedTransaction, RawTransaction]:
        """Gets transaction data.

        :param tx_hash: transaction hash.
        :param simple: (optional) Should the method return a simplified or raw transaction

        :return: transaction data

        :raises: ValueError: if the provided hash is invalid.
        :raises: :class:`KinErrors.ResourceNotFoundError`: if the transaction does not exist.
        :raises: :class:`KinErrors.CantSimplifyError`: if the tx is too complex to simplify
        """
        # TODO: separate to two methods, get_tx_data & get_raw_tx_data
        if not is_valid_transaction_hash(tx_hash):
            raise ValueError('invalid transaction hash: {}'.format(tx_hash))

        try:
            raw_tx = RawTransaction(await self.horizon.transaction(tx_hash))
        except Exception as e:
            raise KinErrors.translate_error(e)

        if simple:
            return SimplifiedTransaction(raw_tx)
        return raw_tx

    async def get_account_tx_history(self, address: str, amount: Optional[int] = 10, descending: Optional[bool] = True,
                                     cursor: Optional[int] = None,
                                     simple: Optional[bool] = True) -> List[Union[SimplifiedTransaction, RawTransaction]]:
        """
        Get the transaction history for a given account.

        :param address: The public address of the account to query
        :param amount: The maximum number of transactions to get
        :param descending: The order of the transactions, True will start from the latest one
        :param cursor: The horizon paging token
        :param simple: Should the returned txs be simplified, if True, complicated txs will be ignored
        :return: A list of transactions
        """

        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))

        if amount <= 0:
            raise ValueError('Limit must be bigger than 0')

        tx_list = []

        requested_amount = amount if amount < MAX_RECORDS_PER_REQUEST else MAX_RECORDS_PER_REQUEST

        horizon_response = await self.horizon.account_transactions(address,
                                                             cursor=cursor, limit=requested_amount,
                                                             order='desc' if descending else 'asc')

        for transaction in horizon_response['_embedded']['records']:
            raw_tx = RawTransaction(transaction)
            if simple:
                try:
                    simple_tx = SimplifiedTransaction(raw_tx)
                    tx_list.append(simple_tx)
                except KinErrors.CantSimplifyError:
                    pass
            else:
                tx_list.append(raw_tx)
            last_cursor = transaction['paging_token']

        remaining_txs = amount - len(tx_list)
        # if we got all the txs that we wanted, or there are no more txs
        # TODO: paging does not work DP-370
        if remaining_txs <= 0 or len(horizon_response['_embedded']['records']) < amount:
            return tx_list
        # If there are anymore transactions, recursively get the next transaction page
        return tx_list.extend(await self.get_account_tx_history(address, remaining_txs, descending, last_cursor, simple))

    async def friendbot(self, address: str) -> str:
        """
        Use the friendbot service to create and fund an account

        :param address: The address to create and fund

        :return: the hash of the friendbot transaction

        :raises ValueError: if no friendbot service was provided
        :raises KinErrors.StellarAddressInvalidError: if the address is invalid
        :raises KinErrors.AccountExistsError if the account already exists
        :raises KinErrors.FriendbotError If the friendbot request failed
        """

        if self.environment.friendbot_url is None:
            raise ValueError("No friendbot service was configured for this client's environments")

        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))
        if await self.does_account_exists(address):
            raise KinErrors.AccountExistsError(address)

        response = await self.horizon._session.get(self.environment.friendbot_url, params={'addr': address})
        if response.status == 200:
            return (await response.json(encoding='utf-8'))['hash']
        else:
            raise KinErrors.FriendbotError(response.status, await response.text(encoding='utf-8'))

    async def friendbot_fund(self, address: str, amount: float) -> str:
        """
        Use the friendbot service to create and fund an account

        :param address: The address to create and fund
        :param amount: How much kin to request from the friendbot

        :return: the hash of the friendbot transaction

        :raises ValueError: if no friendbot service was provided
        :raises ValueError: If the amount requested is invalid
        :raises KinErrors.StellarAddressInvalidError: if the address is invalid
        :raises KinErrors.AccountNotFoundError: if the account does not exist
        :raises KinErrors.FriendbotError If the friendbot request failed
        """

        if self.environment.friendbot_url is None:
            raise ValueError("No friendbot service was configured for this client's environments")

        if not 0 < amount <= 10000:
            raise ValueError("Invalid amount for friendbot request")
        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))
        if not await self.does_account_exists(address):
            raise KinErrors.AccountNotFoundError(address)

        response = await self.horizon._session.get(self.environment.friendbot_url + '/fund', params={'addr': address, 'amount':amount})
        if response.status == 200:
            return (await response.json(encoding='utf-8'))['hash']
        else:
            raise KinErrors.FriendbotError(response.status, await response.text(encoding='utf-8'))

    def monitor_account_payments(self, address: str, timeout: Optional[float] = None) -> AsyncGenerator[SimplifiedTransaction, None]:
        """Monitor KIN payment transactions related to the account identified by provided address.

        :param str address: the address of the account to query.
        :param timeout: How long to wait for each event

        :raises: ValueError: if the address is in the wrong format
        :raises: asyncio.TimeoutError: If too much time has passed between events (only if "timeout" is set)
        """
        return single_monitor(self, address, timeout=timeout)

    def monitor_accounts_payments(self, addresses: set, timeout: Optional[float] = None) -> AsyncGenerator[SimplifiedTransaction, None]:
        """Monitor KIN payment transactions related to multiple accounts

        :param addresses: the addresses of the accounts to query.
        :param timeout: How long to wait for each event

        :raises: asyncio.TimeoutError: If too much time has passed between events (only if "timeout" is set)
        """
        return multi_monitor(self, addresses, timeout=timeout)



"""Contains classes and methods related to transactions and operations"""
from hashlib import sha256
from binascii import hexlify
import base64

from enum import Enum
from kin_base.stellarxdr import Xdr
from kin_base.transaction import Transaction as BaseTransaction
from kin_base.transaction_envelope import TransactionEnvelope as BaseEnvelop
from kin_base.memo import TextMemo, NoneMemo
from kin_base.operation import Payment, CreateAccount

from .errors import CantSimplifyError
from .config import MEMO_TEMPLATE

from typing import Union, Optional

# This is needed in order to calculate transaction hash.
# It is the xdr representation of kin_base.XDR.const.ENVELOP_TYPE_TX (2)
PACKED_ENVELOP_TYPE = b'\x00\x00\x00\x02'
NATIVE_ASSET_TYPE = 'native'


class RawTransaction:
    def __init__(self, horizon_tx_response: dict):
        """
        Class to hold raw info about a transaction

        :param horizon_tx_response: the json response from an horizon query
        """
        # Network_id is left as '' since we override the hash anyway
        self.tx = decode_transaction(horizon_tx_response['envelope_xdr'], network_id='', simple=False)
        self.timestamp = horizon_tx_response['created_at']
        self.hash = horizon_tx_response['hash']


class SimplifiedTransaction:
    def __init__(self, raw_tx: RawTransaction):
        """
        Class to hold simplified info about a transaction

        :param raw_tx: The raw transaction object to simplify
        """
        self.id = raw_tx.hash
        self.timestamp = raw_tx.timestamp

        # If the memo is not a text/none memo
        if not isinstance(raw_tx.tx.memo, (TextMemo, NoneMemo)):
            raise CantSimplifyError('Cant simplify tx with memo type: {}'.format(type(raw_tx.tx.memo)))
        self.memo = None if isinstance(raw_tx.tx.memo, NoneMemo) \
            else raw_tx.tx.memo.text.decode()  # will be none if the there is no memo

        if len(raw_tx.tx.operations) > 1:
            raise CantSimplifyError('Cant simplify tx with {} operations'.format(len(raw_tx.tx.operations)))
        self.operation = SimplifiedOperation(raw_tx.tx.operations[0])

        # Override tx source with operation source if it exists.
        self.source = raw_tx.tx.operations[0].source or raw_tx.tx.source.decode()


class SimplifiedOperation:
    def __init__(self, op_data: Union[CreateAccount, Payment]):
        """
        Class to hold simplified info about a operation

        :param op_data: Operation to simplify
        """
        if isinstance(op_data, Payment):
            # Raise error if its not a KIN payment
            if op_data.asset.type != NATIVE_ASSET_TYPE:
                raise CantSimplifyError('Cant simplify operation with asset {} issued by {}'.
                                        format(op_data.asset.code, op_data.asset.issuer))

            self.amount = float(op_data.amount)
            self.destination = op_data.destination
            self.type = OperationTypes.PAYMENT

        elif isinstance(op_data, CreateAccount):
            self.destination = op_data.destination
            self.starting_balance = float(op_data.starting_balance)
            self.type = OperationTypes.CREATE_ACCOUNT
        else:
            raise CantSimplifyError('Cant simplify operation with {} operation'.format(op_data.type))


class OperationTypes(Enum):
    """Possible operation types for a simple operation"""

    PAYMENT = 1
    CREATE_ACCOUNT = 2


def build_memo(app_id: str, memo: Union[str, None]) -> str:
    """
    Build a memo for a tx that fits the pre-defined template

    :param app_id: The app_id to include in the memo
    :param memo: The memo to include
    :return: the finished memo
    """
    finished_memo = MEMO_TEMPLATE.format(app_id)
    if memo is not None:
        finished_memo += memo

    return finished_memo


def decode_transaction(b64_tx: str, network_id: str, simple: Optional[bool] = True) -> Union[SimplifiedTransaction, RawTransaction]:
    """
    Decode a base64 transaction envelop

    :param b64_tx: a transaction envelop encoded in base64
    :param simple: should the tx be simplified
    :param network_id: the network_id for the transaction
    :return: The transaction

    :raises: KinErrors.CantSimplifyError: if the tx cannot be simplified
    """
    unpacker = Xdr.StellarXDRUnpacker(base64.b64decode(b64_tx))
    envelop = unpacker.unpack_TransactionEnvelope()
    envelop.tx = BaseTransaction.from_xdr_object(envelop.tx)
    passphrase_hash = sha256(network_id.encode()).digest()
    base_tx = BaseEnvelop.from_xdr(b64_tx).tx

    envelop.hash = calculate_tx_hash(base_tx, passphrase_hash)

    # Time cannot be extracted from the envelop
    envelop.timestamp = None
    if simple:
        return SimplifiedTransaction(envelop)
    return envelop.tx


def calculate_tx_hash(tx: BaseTransaction, network_passphrase_hash: bytes) -> str:
    """
    Calculate a tx hash.

    A tx hash is a sha256 hash of:
    1. A sha256 hash of the network_id +
    2. The xdr representation of ENVELOP_TYPE_TX +
    3. The xdr representation of the transaction

    :param tx: The builder's transaction object
    :param network_passphrase_hash: The network passphrase hash
    :return: The hex encoded transaction hash
    """
    # Pack the transaction to xdr
    packer = Xdr.StellarXDRPacker()
    packer.pack_Transaction(tx.to_xdr_object())
    packed_tx = packer.get_buffer()
    return hexlify(sha256(network_passphrase_hash + PACKED_ENVELOP_TYPE + packed_tx).digest()).decode()



"""Contains classes and methods related to channels"""

import sys
import random
from asyncio.queues import Queue as queue
from enum import Enum

from typing import List, Optional

# Python 3.6 didnt support asynccontextmanager, so the kin-sdk installs a backport for it
if sys.version_info.minor == 6:
    from async_generator import asynccontextmanager
else:
    from contextlib import asynccontextmanager


class ChannelManager:
    """Provide useful methods to interact with the underlying ChannelPool"""

    def __init__(self, channel_seeds: List[str]):
        """
        Crete a channel manager instance

        :param channel_seeds: The seeds of the channels to use
        """
        self.channel_pool = ChannelPool(channel_seeds)

    @asynccontextmanager
    async def get_channel(self) -> str:
        """
        Get an available channel

        :return a free channel seed

        """
        channel = await self.channel_pool.get()

        try:
            yield channel
        finally:
            if self.channel_pool._queue[channel] != ChannelStatuses.UNDERFUNDED:
                await self.put_channel(channel)

    async def put_channel(self, channel) -> None:
        """
        Set a channel status back to FREE

        :param str channel: the channel to set back to FREE
        """
        await self.channel_pool.put(channel)

    def get_status(self, verbose: Optional[bool] = False) -> dict:
        """
        Return the current status of the channel manager

        :param verbose: Include all channel seeds and their statuses in the response
        :return: The status of the channel manager
        """
        free_channels = len(self.channel_pool.get_free_channels())
        status = {
            'total_channels': len(self.channel_pool._queue),
            'free_channels': free_channels,
            'non_free_channels': len(self.channel_pool._queue) - free_channels
        }
        if verbose:
            status['channels'] = self.channel_pool._queue

        return status


class ChannelStatuses(str, Enum):
    """Contains possible statuses for channels"""

    # subclass str to be able to serialize to json
    FREE = 'free'
    TAKEN = 'taken'
    UNDERFUNDED = 'underfunded'


class ChannelPool(queue):
    """
    An async queue that sets a member's status instead of pulling it in/out of the queue.
    This queue gets members randomly when 'get' is used, as opposed to always get the last member.
    """
    def __init__(self, channels_seeds):
        """
        Create an instance of ChannelPool
        :param list[str] channels_seeds: The seeds to be put in the queue
        """
        # Init base queue
        super(ChannelPool, self).__init__(len(channels_seeds))
        # Change queue from a 'deque' object to a dict full of free channels
        self._queue = {channel: ChannelStatuses.FREE for channel in channels_seeds}

    def _get(self) -> str:
        """
        Randomly get an available free channel from the dict

        :return: a channel seed
        """
        # Get a list of all free channels
        free_channels = self.get_free_channels()
        # Select a random free channel
        selected_channel = random.choice(free_channels)
        # Change channel state to taken
        self._queue[selected_channel] = ChannelStatuses.TAKEN
        return selected_channel

    def _put(self, channel: str) -> None:
        """
        Change a channel status back to FREE

        :param str channel: the channel seed
        """
        # Change channel state to free
        self._queue[channel] = ChannelStatuses.FREE

    def qsize(self) -> int:
        """
        Counts free channels in the queue

        :return: amount of free channels in the queue
        """
        return len(self.get_free_channels())

    def empty(self) -> bool:
        """Used to check if the queue is empty"""
        return len(self.get_free_channels()) == 0

    def get_free_channels(self) -> List[str]:
        """Get a list of channels with "FREE" status"""
        return [channel for channel, status in self._queue.items() if status == ChannelStatuses.FREE]




"""Contains the KinAccount and AccountStatus classes."""

import re
import json

from kin_base import Builder
from kin_base.network import NETWORKS
from kin_base.transaction_envelope import TransactionEnvelope

from .blockchain.keypair import Keypair
from .blockchain.channel_manager import ChannelManager, ChannelStatuses
from . import errors as KinErrors
from .transactions import build_memo, RawTransaction, SimplifiedTransaction
from .blockchain.errors import TransactionResultCode, HorizonErrorType
from .config import APP_ID_REGEX, KIN_DECIMAL_PRECISION
from .blockchain.utils import is_valid_address, is_valid_secret_key
from .blockchain.horizon_models import AccountData

from typing import List, Optional, Union, AsyncGenerator
import logging

logger = logging.getLogger(__name__)


class KinAccount:
    """Account class to perform authenticated actions on the blockchain"""

    def __init__(self, seed, client, channel_seeds, app_id):
        # Set the internal client
        self._client = client

        # Set the app_id
        self.app_id = app_id

        # Verify the app_id is ok
        if re.match(APP_ID_REGEX, app_id) is None:
            raise ValueError('invalid app id: {}'.format(app_id))

        # Set keypair
        self.keypair = Keypair(seed)

        if channel_seeds is not None:
            # Use given channels
            self.channel_seeds = channel_seeds
            for channel_seed in self.channel_seeds:
                if not is_valid_secret_key(channel_seed):
                    raise KinErrors.StellarSecretInvalidError
        else:
            # Use the base account as the only channel
            self.channel_seeds = [seed]

        self.channel_manager = ChannelManager(self.channel_seeds)

    def get_public_address(self):
        """Return this KinAccount's public address"""
        return self.keypair.public_address

    async def get_balance(self) -> float:
        """
        Get the KIN balance of this KinAccount

        :return: the kin balance

        :raises: KinErrors.AccountNotFoundError if the account does not exist.
        """
        return await self._client.get_account_balance(self.keypair.public_address)

    async def get_data(self) -> AccountData:
        """
        Gets this KinAccount's data

        :return: account data

        :raises: KinErrors.AccountNotFoundError if the account does not exist.
        """
        return await self._client.get_account_data(self.keypair.public_address)

    async def get_status(self, verbose: Optional[bool] = False) -> dict:
        """
        Get the config and status of this KinAccount object

        :param verbose: Should the channels status be verbose
        :return: The config and status of this KinAccount object
        :rtype dict
        """
        account_status = {
            'app_id': self.app_id,
            'public_address': self.get_public_address(),
            'balance': await self.get_balance(),
            'channels': self.channel_manager.get_status(verbose)
        }
        total_status = {
            'client': await self._client.get_config(),
            'account': account_status
        }

        return total_status

    async def get_transaction_history(self, amount: Optional[int] = 10, descending: Optional[bool] = True,
                                      cursor: Optional[int] = None,
                                      simple: Optional[bool] = True) -> List[Union[SimplifiedTransaction, RawTransaction]]:
        """
        Get the transaction history for this kin account

        :param amount: The maximum number of transactions to get
        :param descending: The order of the transactions, True will start from the latest one
        :param cursor: The horizon paging token
        :param simple: Should the returned txs be simplified, if True, complicated txs will be ignored
        :return: A list of transactions
        """

        return await self._client.get_account_tx_history(self.get_public_address(),
                                                         amount=amount,
                                                         descending=descending,
                                                         cursor=cursor,
                                                         simple=simple)

    def get_transaction_builder(self, fee: int) -> Builder:
        """
        Get a transaction builder using this account

        :param fee: The fee that will be used for the transaction
        """
        return Builder(horizon=self._client.horizon,
                       network_name=self._client.environment.name,
                       fee=fee,
                       secret=self.keypair.secret_seed)

    async def create_account(self, address: str, starting_balance: Union[float, str], fee: int,
                             memo_text: Optional[str] = None) -> str:
        """Create an account identified by the provided address.

        :param address: the address of the account to create.
        :param starting_balance: the starting KIN balance of the account.
        :param memo_text: (optional) a text to put into transaction memo, up to MEMO_CAP chars.
        :param fee: fee to be deducted for the tx

        :return: the hash of the transaction

        :raises: KinErrors.StellarAddressInvalidError: if the provided address has a wrong format.
        :raises: KinErrors.AccountExistsError if the account already exists.
        :raises: KinErrors.NotValidParamError if the memo is longer than MEMO_CAP characters
        :raises: KinErrors.NotValidParamError: if the amount is too precise
        :raises: KinErrors.NotValidParamError: if the fee is not valid
        """
        builder = self.build_create_account(address, starting_balance, fee, memo_text)

        async with self.channel_manager.get_channel() as channel:
            await builder.set_channel(channel)
            builder.sign(channel)
            # Also sign with the root account if a different channel was used
            if builder.address != self.keypair.public_address:
                builder.sign(self.keypair.secret_seed)
            return await self.submit_transaction(builder)

    async def send_kin(self, address: str, amount: Union[float, str], fee: int,
                       memo_text: Optional[str] = None) -> str:
        """Send KIN to the account identified by the provided address.

        :param address: the account to send KIN to.
        :param amount: the amount of KIN to send.
        :param memo_text: (optional) a text to put into transaction memo.
        :param fee: fee to be deducted

        :return: the hash of the transaction

        :raises: KinErrors.StellarAddressInvalidError: if the provided address has a wrong format.
        :raises: ValueError: if the amount is not positive.
        :raises: KinErrors.NotValidParamError: if the amount is too precise
        :raises: KinErrors.AccountNotFoundError if the account does not exist.
        :raises: KinErrors.LowBalanceError if there is not enough KIN to send and pay transaction fee.
        :raises: KinErrors.NotValidParamError if the memo is longer than MEMO_CAP characters
        :raises: KinErrors.NotValidParamError: if the fee is not valid
        """
        builder = self.build_send_kin(address, amount, fee, memo_text)
        async with self.channel_manager.get_channel() as channel:
            await builder.set_channel(channel)
            builder.sign(channel)
            # Also sign with the root account if a different channel was used
            if builder.address != self.keypair.public_address:
                builder.sign(self.keypair.secret_seed)
            return await self.submit_transaction(builder)

    def build_create_account(self, address: str, starting_balance: Union[float, str], fee: int,
                          memo_text: Optional[str] = None) -> Builder:
        """Build a tx that will create an account identified by the provided address.
        
        :param address: the address of the account to create.
        :param starting_balance: the starting XLM balance of the account.
        :param memo_text: (optional) a text to put into transaction memo, up to MEMO_CAP chars.
        :param fee: fee to be deducted for the tx

        :return: a transaction builder object

        :raises: KinErrors.StellarAddressInvalidError: if the supplied address has a wrong format.
        """
        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))

        if float(starting_balance) < 0:
            raise ValueError('Starting balance : {} cant be negative'.format(starting_balance))

        # Build the transaction.
        builder = self.get_transaction_builder(fee)
        builder.add_text_memo(build_memo(self.app_id, memo_text))
        builder.append_create_account_op(address, str(starting_balance), source=self.keypair.public_address)
        return builder

    def build_send_kin(self, address: str, amount: Union[float, str], fee: int,
                       memo_text: Optional[str] = None) -> Builder:
        """Build a tx to send KIN to the account identified by the provided address.

        :param address: the account to send asset to.
        :param amount: the KIN amount to send.
        :param memo_text: (optional) a text to put into transaction memo.
        :param fee: fee to be deducted for the tx

        :return: a transaction builder

        :raises: KinErrors.StellarAddressInvalidError: if the provided address has a wrong format.
        :raises: ValueError: if the amount is not positive.
        """

        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))

        if float(amount) <= 0:
            raise ValueError('Amount : {} must be positive'.format(amount))

        builder = self.get_transaction_builder(fee)
        builder.add_text_memo(build_memo(self.app_id, memo_text))
        builder.append_payment_op(address, str(amount), source=self.keypair.public_address)
        return builder

    async def submit_transaction(self, tx_builder: Builder) -> str:
        """
        Submit a transaction to the blockchain.

        :param kin.Builder tx_builder: The transaction builder

        :return: The hash of the transaction.
        :rtype: str
        """
        try:
            return (await tx_builder.submit())['hash']
        # If the channel is out of KIN, top it up and try again
        except KinErrors.HorizonError as e:
            logger.warning('send transaction error with channel {}: {}'.format(tx_builder.address, str(e)))
            if e.type == HorizonErrorType.TRANSACTION_FAILED \
                    and e.extras['result_codes']['transaction'] == TransactionResultCode.INSUFFICIENT_BALANCE:

                self.channel_manager.channel_pool._queue[tx_builder.address] = ChannelStatuses.UNDERFUNDED
                await self._top_up(tx_builder.address)
                self.channel_manager.channel_pool._queue[tx_builder.address] = ChannelStatuses.TAKEN

                # Insufficient balance is a "fast-fail", the sequence number doesn't increment
                # so there is no need to build the transaction again
                await self.submit_transaction(tx_builder)
            else:
                raise KinErrors.translate_error(e)

    def monitor_payments(self, timeout: Optional[float] = None) -> AsyncGenerator[SimplifiedTransaction, None]:
        """Monitor KIN payment transactions related to this account

        :param timeout: How long to wait for each event

        :raises: ValueError: if the address is in the wrong format
        :raises: asyncio.TimeoutError: If too much time has passed between events (only if "timeout" is set)
        """
        return self._client.monitor_account_payments(self.keypair.public_address, timeout)

    def whitelist_transaction(self, payload: Union[str, dict]) -> str:
        """
        Sign on a transaction to whitelist it

        :param payload: the json received from the client
        :return: a signed transaction encoded as base64
        """

        # load the object from the json
        if not isinstance(payload, dict):
            payload = json.loads(payload)

        # If the network the client is using is different from the one we are using
        if self._client.environment.passphrase != payload['network_id']:
            raise KinErrors.WrongNetworkError()

        # The android stellar sdk spells 'tx-envelope' as 'envelop'
        payload_envelope = payload.get('envelop',payload.get('envelope'))
        # Decode the transaction, from_xdr actually takes a base64 encoded xdr
        envelope = TransactionEnvelope.from_xdr(payload_envelope)

        # Add the network_id hash to the envelope
        envelope.network_id = self._client.environment.passphrase_hash

        # Get the transaction hash (to sign it)
        tx_hash = envelope.hash_meta()

        # Sign using the hash
        signature = self.keypair.sign(tx_hash)

        # Add the signature to the envelope
        envelope.signatures.append(signature)

        # Pack the signed envelop to xdr the return it encoded as base64
        # xdr() returns a bytestring that needs to be decoded
        return envelope.xdr().decode()

    # Internal methods

    async def _top_up(self, address: str) -> None:
        """
        Top up a channel with the base account.

        :param address: The address to top up
        """

        # TODO: let user config the amount of kin to top up
        min_fee = await self._client.get_minimum_fee()
        builder = self.get_transaction_builder(min_fee)
        builder.append_payment_op(address, str((min_fee / KIN_DECIMAL_PRECISION) * 1000))  # Enough for 1K txs
        await builder.update_sequence()
        builder.sign()
        await builder.submit()



"""Contains helper methods for the kin sdk"""

from hashlib import sha256

from kin_base import Builder

from .client import KinClient
from .blockchain.keypair import Keypair
from .errors import AccountNotFoundError
from .blockchain.environment import Environment

from typing import List


async def create_channels(master_seed: str, environment: Environment, amount: int,
                          starting_balance: float, salt: str) -> List[str]:
    """
    Create HD seeds based on a master seed and salt

    :param master_seed: The master seed that creates the seeds
    :param environment: The blockchain environment to create the seeds on
    :param amount: Number of seeds to create (Up to 100)
    :param starting_balance: Starting balance to create channels with
    :param salt: A string to be used to create the HD seeds
    :return: The list of seeds generated
    """

    async with KinClient(environment) as client:
        base_key = Keypair(master_seed)
        if not await client.does_account_exists(base_key.public_address):
            raise AccountNotFoundError(base_key.public_address)

        fee = await client.get_minimum_fee()

        channels = get_hd_channels(master_seed, salt, amount)

        # Create a builder for the transaction
        builder = Builder(client.horizon, environment.name, fee, master_seed)

        # Find out if this salt+seed combination was ever used to create channels.
        # If so, the user might only be interested in adding channels,
        # so we need to find what seed to start from

        # First check if the last channel exists, if it does, we don't need to create any channel.
        if await client.does_account_exists(Keypair.address_from_seed(channels[-1])):
            return channels

        for index, seed in enumerate(channels):
            if await client.does_account_exists(Keypair.address_from_seed(seed)):
                continue

            # Start creating from the current seed forward
            for channel_seed in channels[index:]:
                builder.append_create_account_op(Keypair.address_from_seed(channel_seed), str(starting_balance))

            await builder.update_sequence()
            builder.sign()
            await builder.submit()
            break

    return channels


def get_hd_channels(master_seed: str, salt: str, amount: int) -> List[str]:
    """
    Get a list of channels generated based on a seed and salt

    :param master_seed: the base seed that created the channels
    :param salt: A string to be used to generate the seeds
    :param amount: Number of seeds to generate (Up to 100)
    :return: The list of seeds generated
    :rtype list[str]
    """

    if amount > 100:
        """
        The sdk's channels are not meant to be shared across multiple instances of the script,
        and a single instance will never even use 100 channels at once.
        This is a limit to stop developers from needlessly creating a huge amount of channels
        """
        raise ValueError('Only up to 100 channels can be created with a specific seed + salt combination')
    hashed_salt = sha256(salt.encode()).hexdigest()

    channels = []
    for index in range(amount):
        # The salt used to generate the
        channel = Keypair.generate_hd_seed(master_seed, hashed_salt + str(index))
        channels.append(channel)

    return channels







"""Contains classes for monitoring the blockchain"""
from .blockchain.utils import is_valid_address
from .transactions import OperationTypes, SimplifiedTransaction, RawTransaction
from .errors import CantSimplifyError, StellarAddressInvalidError

from typing import Optional, AsyncGenerator

import logging

logger = logging.getLogger(__name__)


async def single_monitor(kin_client: 'KinClient', address: str,
                         timeout: Optional[float] = None) -> AsyncGenerator[SimplifiedTransaction, None]:
    """
    Monitors a single account for kin payments

    :param kin_client: a kin client directed to the correct network
    :param address: address to watch
    :param timeout: How long to wait for a new event

    :raises: asyncio.TimeoutError: If too much time has passed between events (only if "timeout" is set)
    """
    if not is_valid_address(address):
        raise StellarAddressInvalidError('invalid address: {}'.format(address))

    sse_client = await kin_client.horizon.account_transactions(address, sse=True, sse_timeout=timeout)

    async for tx in sse_client:
        try:
            tx_data = SimplifiedTransaction(RawTransaction(tx))
        except CantSimplifyError:
            logger.debug("SSE transaction couldn't be simplified: ", tx)
            continue

        if tx_data.operation.type != OperationTypes.PAYMENT:
            logger.debug("Non-payment SSE transaction skipped: ", tx_data)
            continue

        yield tx_data


async def multi_monitor(kin_client: 'KinClient', addresses: set) -> AsyncGenerator[SimplifiedTransaction, None]:
    """
    Monitors a single account for kin payments

    :param kin_client: a kin client directed to the correct network
    :param addresses: set of addresses to watch
    """

    sse_client = await kin_client.horizon.transactions(sse=True)

    async for tx in sse_client:
        try:
            tx_data = SimplifiedTransaction(RawTransaction(tx))
        except CantSimplifyError:
            logger.debug("SSE transaction couldn't be simplified: ", tx)
            continue

        if tx_data.operation.type != OperationTypes.PAYMENT:
            logger.debug("Non-payment SSE transaction skipped: ", tx_data)
            continue

        # Will yield twice if both of these are correct. (someone sent to himself) - which it fine
        if tx_data.source in addresses:
            yield tx_data.source, tx_data
        if tx_data.operation.destination in addresses:
            yield tx_data.operation.destination, tx_data